# Copyright 2019 Red Hat
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import tobiko
from tobiko import config
from tobiko.openstack import glance
from tobiko.openstack import topology
from tobiko.openstack.stacks import _nova
from tobiko.openstack.stacks import _vlan
from tobiko.shell import sh


CONF = config.CONF


class AdvancedImageFixture(glance.UrlGlanceImageFixture):
    """Advanced server image running an HTTP server

    The server has additional installed packages compared to
    the minimal one:

    - iperf3
    - ping
    - ncat
    - nginx

    The image will also have below running services:

    - nginx HTTP server listening on TCP port 80
    - iperf3 server listening on TCP port 5201
    """
    image_url = CONF.tobiko.advanced_vm.image_url
    image_name = CONF.tobiko.advanced_vm.image_name
    disk_format = CONF.tobiko.advanced_vm.disk_format or "qcow2"
    container_format = CONF.tobiko.advanced_vm.container_format or "bare"
    username = CONF.tobiko.advanced_vm.username or 'cloud-user'
    connection_timeout = CONF.tobiko.nova.advanced_vm_connection_timeout
    disabled_algorithms = CONF.tobiko.advanced_vm.disabled_algorithms
    is_reachable_timeout = CONF.tobiko.nova.advanced_vm_is_reachable_timeout

    # port of running HTTP server
    http_port = 80
    # port of running Iperf3 server
    iperf3_port = 5201

    iperf3_service_name = "iperf3-server.service"

    @property
    def vlan_id(self) -> int:
        return tobiko.tobiko_config().neutron.vlan_id

    @property
    def vlan_device(self) -> str:
        return f'vlan{self.vlan_id}'


class AdvancedFlavorStackFixture(_nova.FlavorStackFixture):
    ram = 256
    swap = 512


class AdvancedServerStackFixture(_nova.CloudInitServerStackFixture,
                                 _vlan.VlanServerStackFixture):
    """Advanced server running an HTTP server

    The server has additional commands compared to the minimal one:

    - iperf3
    - ping
    """

    #: Glance image used to create a Nova server instance
    image_fixture = tobiko.required_fixture(AdvancedImageFixture)
    #: Flavor used to create a Nova server instance
    flavor_stack = tobiko.required_fixture(AdvancedFlavorStackFixture)

    @property
    def is_reachable_timeout(self) -> tobiko.Seconds:
        return self.image_fixture.is_reachable_timeout

    # port of running HTTP server
    @property
    def http_port(self) -> int:
        return self.image_fixture.http_port

    # port of running Iperf3 server
    @property
    def iperf3_port(self) -> int:
        return self.image_fixture.iperf3_port

    @property
    def iperf3_service_name(self) -> str:
        return self.image_fixture.iperf3_service_name

    def wait_for_iperf3_server(self,
                               timeout: tobiko.Seconds = None,
                               interval: tobiko.Seconds = None):
        return sh.wait_for_active_systemd_units(self.iperf3_service_name,
                                                timeout=timeout,
                                                interval=interval,
                                                ssh_client=self.ssh_client)

    @property
    def vlan_id(self) -> int:
        return self.image_fixture.vlan_id

    @property
    def vlan_device(self) -> str:
        return self.image_fixture.vlan_device


@topology.skip_unless_osp_version('17.0', lower=True)
class AdvancedExternalServerStackFixture(AdvancedServerStackFixture,
                                         _nova.ExternalServerStackFixture):
    """Advanced server with port on special external network
    """


class AdvancedPeerServerStackFixture(AdvancedServerStackFixture,
                                     _nova.PeerServerStackFixture):
    #: Peer server used to reach this one
    peer_stack = tobiko.required_fixture(AdvancedServerStackFixture)



# Copyright (c) 2024 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import testtools
from oslo_log import log

import tobiko
from tobiko.openstack import topology
from tobiko.podified import _topology
from tobiko.shell import sh

LOG = log.getLogger(__name__)


class KpatchCommandNotFoundError(tobiko.TobikoException):
    message = "kpatch command not found on compute node"


class KpatchCommandFailedError(tobiko.TobikoException):
    message = "kpatch command failed on compute node"


class KpatchNotLoadedError(tobiko.TobikoException):
    message = "no kpatches loaded on compute node"


@_topology.skip_if_not_podified
class KpatchVerificationTest(testtools.TestCase):
    """Test to verify kpatches are applied on compute nodes after update"""

    def test_kpatches_applied_on_compute_nodes(self):
        """Verify that kpatches are loaded on all compute nodes"""
        compute_nodes = topology.list_openstack_nodes(group='compute')

        if not compute_nodes:
            tobiko.skip_test(
                "No EDPM compute nodes found in podified topology")

        LOG.info("Checking kpatches on {} compute nodes".format(
            len(compute_nodes)))

        failed_nodes = []

        for node in compute_nodes:
            if not isinstance(node, _topology.EdpmNode):
                LOG.warning(f"Skipping non-EDPM node: {node.name}")
                continue

            try:
                # Execute kpatch list command
                result = sh.execute('kpatch list',
                                    ssh_client=node.ssh_client,
                                    expect_exit_status=None)

                # Check if kpatch command not found
                if result.exit_status == 127:
                    raise KpatchCommandNotFoundError(
                        f"kpatch command not found on {node.name}")

                # Check if kpatch command failed
                if result.exit_status != 0:
                    raise KpatchCommandFailedError(
                        "kpatch command failed on {} "
                        "(exit code: {}): {}".format(
                            node.name, result.exit_status, result.stderr))

                # Check if no output (no kpatches loaded)
                output = result.stdout.strip()
                if not output:
                    raise KpatchNotLoadedError(
                        f"No kpatches loaded on {node.name}")

                LOG.info("Kpatches on {}:\n{}".format(node.name, output))

            except tobiko.TobikoException:
                # Re-raise tobiko exceptions as-is
                raise
            except Exception as ex:
                LOG.exception(f"Failed to check kpatches on {node.name}")
                failed_nodes.append("Exception on {}: {}".format(
                    node.name, str(ex)))

        # This should only be reached if there were non-tobiko exceptions
        if failed_nodes:
            failure_message = "Kpatch verification failed:\n" + "\n".join(
                "  - {}".format(failure) for failure in failed_nodes)
            raise AssertionError(failure_message)



# Copyright (c) 2019 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from collections import abc
import contextlib
import getpass
import os
import subprocess
import time
import threading
import typing

import netaddr
import testtools
from oslo_log import log
from oslo_utils.secretutils import md5
import paramiko
from paramiko import common

import tobiko
from tobiko.shell.ssh import _config
from tobiko.shell.ssh import _command


LOG = log.getLogger(__name__)


def get_fingerprint(self):
    """Patch paramiko

    This method needs to be patched to allow paramiko to work under FIPS.
    Until the patch to do this merges, patch paramiko here.

    TODO(eolivare) Remove this when paramiko is patched.
    See https://github.com/paramiko/paramiko/pull/1928
    """
    return md5(self.asbytes(), usedforsecurity=False).digest()


# mypy: disable-error-code="assignment"
paramiko.pkey.PKey.get_fingerprint = get_fingerprint


def valid_hostname(value):
    hostname = str(value)
    if not hostname:
        message = "Invalid hostname: {!r}".format(hostname)
        raise ValueError(message)
    return hostname


def valid_port(value):
    port = int(value)
    if port <= 0 or port > 65535:
        message = "Invalid port number: {!r}".format(port)
        raise ValueError(message)
    return port


def valid_path(value):
    return os.path.abspath(os.path.expanduser(value))


def positive_float(value):
    value = float(value)
    if value <= 0.:
        message = "{!r} is not positive".format(value)
        raise ValueError(message)
    return value


def positive_int(value):
    value = int(value)
    if value <= 0:
        message = "{!r} is not positive".format(value)
        raise ValueError(message)
    return value


def get_key_filename(value):
    if isinstance(value, str):
        value = [value]
    key_filename = [tobiko.tobiko_config_path(v) for v in value]
    return [f
            for f in key_filename
            if os.path.isfile(f)]


SSH_CONNECT_PARAMETERS = {
    #: The server to connect to
    'hostname': valid_hostname,

    #: The server port to connect to
    'port': valid_port,

    #: The username to authenticate as (defaults to the current local username)
    'username': str,

    #: Used for password authentication; is also used for private key
    #: decryption if passphrase is not given
    'password': str,

    #: Used for decrypting private keys
    'passphrase': str,

    #: The filename, or list of filenames, of optional private key(s) and/or
    #: certs to try for authentication
    'key_filename': get_key_filename,

    #: An optional timeout (in seconds) for the TCP connect
    'timeout': positive_float,

    #: Set to False to disable connecting to the SSH agent
    'allow_agent': bool,

    #: Set to False to disable searching for discoverable private key files in
    #: ~/.ssh/
    'look_for_keys': bool,

    #: Set to True to turn on compression
    'compress': bool,

    #: True if you want to use GSS-API authentication
    'gss_auth': bool,

    #: Perform GSS-API Key Exchange and user authentication
    'gss_kex': bool,

    #: Delegate GSS-API client credentials or not
    'gss_deleg_creds': bool,

    #: The targets name in the kerberos database. default: hostname
    'gss_host': str,

    #: Indicates whether or not the DNS is trusted to securely canonicalize the
    #: name of the host being connected to (default True).
    'gss_trust_dns': bool,

    #: An optional timeout (in seconds) to wait for the SSH banner to be
    #: presented.
    'banner_timeout': positive_float,

    #: An optional timeout (in seconds) to wait for an authentication response
    'auth_timeout': positive_float,

    #: Number of connection attempts to be tried before timeout
    'connection_attempts': positive_int,

    #: Minimum amount of time to wait between two connection attempts
    'connection_interval': positive_float,

    #: Amount of time before timing our connection attempt
    'connection_timeout': positive_float,

    #: Command to be executed to open proxy sock
    'proxy_command': str,

    #: Allow to disable some algorithms for old servers
    'disabled_algorithms': dict
}


def gather_ssh_connect_parameters(source=None, destination=None, schema=None,
                                  remove_from_schema=False, **kwargs):
    if schema is None:
        assert not remove_from_schema
        schema = SSH_CONNECT_PARAMETERS
    parameters = {}

    if source:
        # gather from object
        if isinstance(source, abc.Mapping):
            parameters.update(_items_from_mapping(mapping=source,
                                                  schema=schema))
        else:
            parameters.update(_items_from_object(obj=source,
                                                 schema=schema))

    if kwargs:
        # gather from kwargs
        parameters.update(_items_from_mapping(mapping=kwargs, schema=schema))
        kwargs = exclude_mapping_items(kwargs, schema)
        if kwargs:
            message = "Invalid SSH connection parameters: {!r}".format(kwargs)
            raise ValueError(message)

    if remove_from_schema and parameters:
        # update schema
        for name in parameters:
            del schema[name]

    if destination is not None:
        destination.update(parameters)
    return parameters


def _items_from_mapping(mapping, schema):
    for name, init in schema.items():
        value = mapping.get(name)
        if value is not None:
            yield name, init(value)


def _items_from_object(obj, schema):
    for name, init in schema.items():
        value = getattr(obj, name, None)
        if value is not None:
            yield name, init(value)


def exclude_mapping_items(mapping, exclude):
    """Exclude parameters that are already in target dictionary"""
    return {key: value
            for key, value in mapping.items()
            if key not in exclude}


class SSHConnectFailure(tobiko.TobikoException):
    message = "Failed to login to {login}\n{cause}"


class SSHClientFixture(tobiko.SharedFixture):

    host = None
    port = 22
    username = getpass.getuser()

    client = None

    default = tobiko.required_fixture(_config.SSHDefaultConfigFixture)
    config_files = None
    host_config = None
    global_host_config = None

    proxy_client = None
    proxy_sock = None
    connect_parameters = None
    schema = SSH_CONNECT_PARAMETERS

    def __init__(self, host=None, proxy_client=None, host_config=None,
                 config_files=None, schema=None, **kwargs):
        super(SSHClientFixture, self).__init__()
        if host:
            self.host = host
        if proxy_client:
            self.proxy_client = proxy_client
        if host_config:
            self.host_config = host_config
        if config_files:
            self.config_files = config_files

        self.schema = schema = dict(schema or self.schema)
        self._connect_parameters = gather_ssh_connect_parameters(
            schema=schema, **kwargs)
        self._forwarders = []

    def setup_fixture(self):
        self.setup_connect_parameters()
        self.setup_ssh_client()

    def setup_global_host_config(self):
        self.global_host_config = config = _config.ssh_host_config(
            host=self.host, config_files=self.config_files)
        return config

    def setup_connect_parameters(self):
        """Fill connect parameters dict

        Get parameters values from below sources:
        - parameters passed to class constructor
        - parameters got from ~/.ssh/config and tobiko.conf
        - parameters got from fixture object attributes
        """
        self.setup_global_host_config()
        self.connect_parameters = parameters = self.get_connect_parameters()
        return parameters

    def get_connect_parameters(self, schema=None):
        schema = dict(schema or self.schema)
        parameters = {}
        self.gather_initial_connect_parameters(
            destination=parameters, schema=schema, remove_from_schema=True)
        self.gather_global_host_config_connect_parameters(
            destination=parameters, schema=schema, remove_from_schema=True)
        self.gather_host_config_connect_parameters(
            destination=parameters, schema=schema, remove_from_schema=True)
        self.gather_default_connect_parameters(
            destination=parameters, schema=schema, remove_from_schema=True)
        return parameters

    def gather_initial_connect_parameters(self, **kwargs):
        if self._connect_parameters:
            gather_ssh_connect_parameters(
                source=self._connect_parameters, **kwargs)

    def gather_host_config_connect_parameters(self, **kwargs):
        if self.host_config:
            gather_ssh_connect_parameters(
                source=self.host_config.connect_parameters, **kwargs)

    def gather_global_host_config_connect_parameters(self, **kwargs):
        if self.global_host_config:
            gather_ssh_connect_parameters(
                source=self.global_host_config.connect_parameters, **kwargs)

    def gather_default_connect_parameters(self, **kwargs):
        return gather_ssh_connect_parameters(source=self, **kwargs)

    def setup_ssh_client(self):
        self.client, self.proxy_sock = ssh_connect(
            proxy_client=self.proxy_client,
            **self.connect_parameters)
        self.addCleanup(self.cleanup_ssh_client)
        if self.proxy_sock:
            self.addCleanup(self.cleanup_proxy_sock)
        for forwarder in self._forwarders:
            self.useFixture(forwarder)

    def cleanup_ssh_client(self):
        client = self.client
        self.client = None
        if client:
            LOG.debug(f"Closing SSH connection... ({self.details})")
            try:
                client.close()
            except Exception:
                LOG.exception("Error closing SSH connection. "
                              f"({self.details})")
            else:
                LOG.debug(f"SSH connection closed. ({self.details})")

    def cleanup_proxy_sock(self):
        proxy_sock = self.proxy_sock
        self.proxy_sock = None
        if proxy_sock:
            LOG.debug(f"Closing SSH proxy sock... ({self.details})")
            try:
                proxy_sock.close()
            except Exception:
                LOG.exception("Error closing proxy socket. "
                              f"({self.details})")
            else:
                LOG.debug(f"SSH proxy sock closed. ({self.details})")

    @contextlib.contextmanager
    def use_connect_parameters(self, **kwargs):
        if kwargs:
            restore_parameters = dict(self._connect_parameters)
            gather_ssh_connect_parameters(schema=self.schema,
                                          destination=self._connect_parameters,
                                          **kwargs)
            self.connect_parameters = self.get_connect_parameters()
        else:
            restore_parameters = None
        try:
            yield
        finally:
            if restore_parameters is not None:
                self._connect_parameters = restore_parameters
                self.connect_parameters = self.get_connect_parameters()

    def connect(self, retry_count: typing.Optional[int] = 2,
                retry_timeout: tobiko.Seconds = None,
                retry_interval: tobiko.Seconds = None,
                **ssh_parameters) -> paramiko.SSHClient:
        """Ensures it is connected to remote SSH server
        """
        with self.use_connect_parameters(**ssh_parameters):
            # This retry is mostly intended to ensure connection is
            # reestablished in case it is lost
            for attempt in tobiko.retry(count=retry_count,
                                        timeout=retry_timeout,
                                        interval=retry_interval):
                LOG.debug(f"Ensuring SSH connection (attempt={attempt})")
                try:
                    # Create a new paramiko.SSHClient() instance and connect it
                    # to the a server using the provided credentials
                    client = tobiko.setup_fixture(self).client
                except Exception:
                    attempt.check_limits()
                    LOG.exception(f"Failed connecting to '{self.login}' "
                                  f"(attempt={attempt})")
                    self.close()
                else:
                    break

        return client

    def close(self):
        """Ensures it is disconnected from remote SSH server
        """
        try:
            tobiko.cleanup_fixture(self)
        except Exception:
            LOG.exception(f"Failed closing SSH connection to '{self.login}'")

    def get_ssh_command(self, host=None, username=None, port=None,
                        command=None, config_files=None, host_config=None,
                        proxy_command=None, key_filename=None, **options):
        connect_parameters = self.setup_connect_parameters()
        host = host or connect_parameters.get('hostname')
        username = username or connect_parameters.get('username')
        port = port or connect_parameters.get('port')
        config_files = config_files or connect_parameters.get('config_files')
        if not host_config:
            _host_config = self.setup_global_host_config()
            if hasattr(_host_config, 'host_config'):
                _host_config = host_config
        key_filename = key_filename or connect_parameters.get('key_filename')
        proxy_command = (proxy_command or
                         connect_parameters.get('proxy_command'))
        if not proxy_command and self.proxy_client:
            proxy_command = self.proxy_client.get_ssh_command()

        return _command.ssh_command(host=host,
                                    username=username,
                                    port=port,
                                    command=command,
                                    config_files=config_files,
                                    host_config=host_config,
                                    proxy_command=proxy_command,
                                    key_filename=key_filename,
                                    **options)

    @property
    def login(self):
        parameters = self.setup_connect_parameters()
        return _command.ssh_login(hostname=parameters['hostname'],
                                  username=parameters['username'],
                                  port=parameters['port'])

    @property
    def hostname(self):
        parameters = self.setup_connect_parameters()
        return parameters['hostname']

    @property
    def details(self):
        return f"login='{self.login}'"

    def open_unix_socket(self,
                         socket_path: str,
                         window_size: int = common.DEFAULT_WINDOW_SIZE,
                         max_packet_size: int = common.DEFAULT_MAX_PACKET_SIZE,
                         timeout: tobiko.Seconds = None) \
            -> paramiko.Channel:
        # pylint: disable=protected-access
        transport: typing.Any = self.connect().get_transport()
        if transport is None or not transport.active:
            raise paramiko.SSHException("SSH session not active")
        timeout = 3600 if timeout is None else timeout
        transport.lock.acquire()
        try:
            window_size = transport._sanitize_window_size(window_size)
            max_packet_size = transport._sanitize_packet_size(max_packet_size)
            chanid = transport._next_channel()

            # Documented here:
            # https://github.com/openssh/openssh-portable/blob/master/PROTOCOL
            m = paramiko.Message()
            m.add_byte(b'Z')
            m.add_string('direct-streamlocal@openssh.com')
            m.add_int(chanid)
            m.add_int(window_size)
            m.add_int(max_packet_size)
            m.add_string(socket_path)
            # Reserved stuff
            m.add_string('')
            m.add_int(0)

            sock: typing.Any = paramiko.Channel(chanid)
            transport._channels.put(chanid, sock)
            transport.channel_events[chanid] = event = threading.Event()
            transport.channels_seen[chanid] = True
            sock._set_transport(transport)
            sock._set_window(window_size, max_packet_size)
        finally:
            transport.lock.release()

        transport._send_user_message(m)
        start_ts = tobiko.time()
        while True:
            event.wait(0.1)
            if not transport.active:
                e = transport.get_exception()
                if e is None:
                    e = paramiko.SSHException("Unable to open channel.")
                raise e
            if event.is_set():
                break
            elif start_ts + timeout < time.time():
                raise paramiko.SSHException("Timeout opening channel.")

        sock = transport._channels.get(chanid)
        if sock is None:
            e = transport.get_exception()
            if e is None:
                e = paramiko.SSHException("Unable to open channel.")
            raise e

        return sock

    def __repr__(self):
        return f"SSHClientFixture <{self.details}>"


UNDEFINED_CLIENT = 'UNDEFINED_CLIENT'


class SSHClientManager(object):

    default = tobiko.required_fixture(_config.SSHDefaultConfigFixture)

    def __init__(self):
        self.clients = {}

    def get_client(self, host, hostname=None, username=None, port=None,
                   proxy_jump=None, host_config=None, config_files=None,
                   proxy_client=None, **connect_parameters) -> \
            SSHClientFixture:
        if isinstance(host, netaddr.IPAddress):
            host = str(host)

        if host_config:
            hostname = hostname or host_config.hostname
            port = port or host_config.port
            username = username or host_config.username

        global_host_config = _config.ssh_host_config(host=host,
                                                     config_files=config_files)
        hostname = hostname or global_host_config.hostname
        port = port or global_host_config.port
        username = username or global_host_config.username

        host_key = hostname, port, username, proxy_jump
        existing_client = self.clients.get(host_key)
        if isinstance(existing_client, SSHClientFixture):
            return existing_client

        # Put a placeholder to avoid infinite recursive lookup
        if existing_client is UNDEFINED_CLIENT:
            raise RuntimeError('Recursive SSH proxy client definition')
        self.clients[host_key] = UNDEFINED_CLIENT

        proxy_client = proxy_client or self.get_proxy_client(
            host=host, proxy_jump=proxy_jump, config_files=config_files)
        self.clients[host_key] = new_client = SSHClientFixture(
            host=host, hostname=hostname, port=port, username=username,
            proxy_client=proxy_client, host_config=host_config,
            **connect_parameters)
        return new_client

    def get_proxy_client(self,
                         host=None,
                         proxy_jump=None,
                         host_config=None,
                         config_files=None):
        if isinstance(proxy_jump, SSHClientFixture):
            return proxy_jump
        host_config = host_config or _config.ssh_host_config(
            host=host, config_files=config_files)
        proxy_jump = host_config.proxy_jump
        return proxy_jump and self.get_client(proxy_jump) or None


CLIENTS = SSHClientManager()


def ssh_client(host, port=None, username=None, proxy_jump=None,
               host_config=None, config_files=None, manager=None,
               **connect_parameters) -> SSHClientFixture:
    manager = manager or CLIENTS
    return manager.get_client(host=host, port=port, username=username,
                              proxy_jump=proxy_jump, host_config=host_config,
                              config_files=config_files,
                              **connect_parameters)


KEY_CLASSES: typing.List[typing.Type[paramiko.PKey]] = [
    paramiko.RSAKey,
    paramiko.DSSKey,
    paramiko.ECDSAKey,
    paramiko.Ed25519Key,
]


def load_private_keys(key_filenames: typing.List[str],
                      password: str = None) \
        -> typing.List[paramiko.PKey]:
    pkeys: typing.List[paramiko.PKey] = []
    for filename in key_filenames:
        if os.path.exists(filename):
            try:
                pkey = load_private_key(filename, password=password)
            except LoadPrivateKeyError as ex:
                LOG.exception(f'Error loading key file: {ex}')
            else:
                pkeys.append(pkey)
        else:
            LOG.debug(f'Key file not found: {filename}')
    return pkeys


class LoadPrivateKeyError(tobiko.TobikoException):
    message = "Unable to load private key from file {filename}"


def load_private_key(filename: str,
                     password: str = None) -> paramiko.PKey:
    errors: typing.List[tobiko.ExceptionInfo] = []
    for key_class in KEY_CLASSES:
        try:
            pkey: paramiko.PKey = key_class.from_private_key_file(
                filename=filename, password=password)
        except (paramiko.SSHException, ValueError):
            errors.append(tobiko.exc_info())
        else:
            LOG.debug(f'Key file loaded: {filename} (key_class={key_class})')
            return pkey

    cause = testtools.MultipleExceptions(*errors)
    raise LoadPrivateKeyError(filename=filename) from cause


def ssh_connect(hostname, username=None, port=None, connection_interval=None,
                connection_attempts=None, connection_timeout=None,
                proxy_command=None, proxy_client=None, key_filename=None,
                password: str = None,
                **parameters):
    client = paramiko.SSHClient()
    client.set_missing_host_key_policy(paramiko.WarningPolicy())
    login = _command.ssh_login(hostname=hostname, username=username, port=port)

    assert isinstance(key_filename, list)
    pkeys = load_private_keys(key_filename, password=password)
    auth_failed: typing.Optional[Exception] = None
    for attempt in tobiko.retry(count=connection_attempts,
                                timeout=connection_timeout,
                                interval=connection_interval,
                                default_count=60,
                                default_timeout=300.,
                                default_interval=5.):
        LOG.debug(f"Logging in to '{login}'...\n"
                  f"  - parameters: {parameters}\n"
                  f"  - attempt: {attempt.details}\n")
        for pkey in pkeys + [None]:  # type: ignore
            succeeded = False
            proxy_sock = ssh_proxy_sock(
                hostname=hostname,
                port=port,
                command=proxy_command,
                client=proxy_client,
                timeout=connection_timeout,
                connection_attempts=1,
                connection_interval=connection_interval)
            try:
                client.connect(hostname=hostname,
                               username=username,
                               port=port,
                               sock=proxy_sock,
                               pkey=pkey,
                               **parameters)
            except paramiko.ssh_exception.AuthenticationException as ex:
                if auth_failed is not None:
                    ex.__cause__ = auth_failed
                auth_failed = ex
                continue
            except Exception as ex:
                LOG.debug(f"Error logging in to '{login}': {ex}", exc_info=1)
                attempt.check_limits()
                break
            else:
                LOG.debug(f"Successfully logged in to '{login}'")
                succeeded = True
                return client, proxy_sock
            finally:
                if not succeeded:
                    try:
                        proxy_sock.close()
                    except Exception:
                        pass
        else:
            if isinstance(auth_failed, Exception):
                raise auth_failed


def ssh_proxy_sock(hostname=None, port=None, command=None, client=None,
                   source_address=None, timeout=None,
                   connection_attempts=None, connection_interval=None):
    if not command:
        if client:
            # I need a command to execute with proxy client
            options = []
            if source_address:
                options += ['-s', str(source_address)]
            command = ['nc'] + options + ['{hostname!s}', '{port!s}']
        elif not command:
            # Proxy sock is not required
            return None

    # Apply connect parameters to proxy command
    if not isinstance(command, str):
        command = subprocess.list2cmdline(command)
    if hostname:
        command = command.format(hostname=hostname, port=(port or 22))
    if client:
        if isinstance(client, SSHClientFixture):
            # Connect to proxy server
            client = client.connect(connection_timeout=timeout,
                                    connection_attempts=connection_attempts,
                                    connection_interval=connection_interval)
        elif not isinstance(client, paramiko.SSHClient):
            message = "Object {!r} is not an SSHClient".format(client)
            raise TypeError(message)

        # Open proxy channel
        LOG.debug("Execute proxy command with proxy client %r: %r",
                  client, command)
        sock = client.get_transport().open_session(timeout=timeout)
        sock.exec_command(command)
    else:
        LOG.debug("Execute proxy command on local host: %r", command)
        sock = paramiko.ProxyCommand(command)

    return sock


def ssh_proxy_client(manager=None,
                     host=None,
                     host_config=None,
                     config_files=None):
    manager = manager or CLIENTS
    return manager.get_proxy_client(host=host,
                                    host_config=host_config,
                                    config_files=config_files)


SSHClientType = typing.Union[None, bool, SSHClientFixture]


def ssh_client_fixture(obj: SSHClientType) -> \
        typing.Optional[SSHClientFixture]:
    if obj is None:
        return ssh_proxy_client()
    if obj is False:
        return None
    if isinstance(obj, SSHClientFixture):
        return obj
    raise TypeError(f"Can't get an SSHClientFixture from objeck {obj}")



# Copyright (c) 2021 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import
from __future__ import division

import json
import os
import typing

import netaddr
from oslo_log import log

import tobiko
from tobiko import config
from tobiko.shell import files
from tobiko.shell.iperf3 import _interface
from tobiko.shell.iperf3 import _parameters
from tobiko.shell import sh
from tobiko.shell import ssh


CONF = config.CONF
LOG = log.getLogger(__name__)


def get_iperf3_logs_filepath(address: typing.Union[str, netaddr.IPAddress],
                             path: str,
                             ssh_client: ssh.SSHClientType = None) -> str:
    final_dir = files.get_home_absolute_filepath(path, ssh_client)
    filename = f'iperf_{address}.log'
    return os.path.join(final_dir, filename)


def get_bandwidth(address: typing.Union[str, netaddr.IPAddress],
                  bitrate: int = None,
                  download: bool = None,
                  port: int = None,
                  protocol: str = None,
                  ssh_client: ssh.SSHClientType = None,
                  timeout: tobiko.Seconds = None) -> float:
    iperf_measures = execute_iperf3_client(address=address,
                                           bitrate=bitrate,
                                           download=download,
                                           port=port,
                                           protocol=protocol,
                                           ssh_client=ssh_client,
                                           timeout=timeout)
    return calculate_bandwith(iperf_measures)


def calculate_bandwith(iperf_measures) -> float:
    # first interval is removed because BW measured during it is not
    # limited - it takes ~ 1 second to traffic shaping algorithm to apply
    # bw limit properly (buffer is empty when traffic starts being sent)
    intervals = iperf_measures['intervals'][1:]
    bits_received = sum([interval['sum']['bytes'] * 8
                         for interval in intervals])
    elapsed_time = sum([interval['sum']['seconds']
                        for interval in intervals])
    # bw in bits per second
    return bits_received / elapsed_time


def execute_iperf3_client(address: typing.Union[str, netaddr.IPAddress],
                          bitrate: int = None,
                          download: bool = None,
                          port: int = None,
                          protocol: str = None,
                          ssh_client: ssh.SSHClientType = None,
                          timeout: tobiko.Seconds = None,
                          interval: int = None,
                          logfile: str = None,
                          run_in_background: bool = False) \
        -> typing.Dict:
    params_timeout: typing.Optional[int] = None
    if run_in_background:
        params_timeout = 0
    elif timeout is not None:
        params_timeout = int(timeout - 0.5)

    # json_stream option should be set when:
    # - tests are run_in_background
    # - tests are executed from a VM instance
    # This way we avoid using that option from the test machine (undercloud,
    # test pod, devstack, ...) where it is not supported
    # iperf3 support this option from version 3.17
    json_stream = run_in_background and (ssh_client is not None)

    parameters = _parameters.iperf3_client_parameters(
        address=address, bitrate=bitrate,
        download=download, port=port, protocol=protocol,
        timeout=params_timeout, interval=interval,
        json_stream=json_stream, logfile=logfile)
    command = _interface.get_iperf3_client_command(parameters)

    # output is a dictionary
    if run_in_background:
        process = sh.process(command, ssh_client=ssh_client)
        process.execute()
        return {}
    output = sh.execute(command,
                        ssh_client=ssh_client,
                        timeout=timeout).stdout
    return json.loads(output)


def execute_iperf3_client_in_background(
        address: typing.Union[str, netaddr.IPAddress],  # noqa; pylint: disable=W0613
        bitrate: int = None,
        download: bool = None,
        port: int = None,
        protocol: str = None,
        ssh_client: ssh.SSHClientType = None,
        iperf3_server_ssh_client: ssh.SSHClientType = None,
        output_dir: str = 'tobiko_iperf_results',
        **kwargs) -> None:
    output_path = get_iperf3_logs_filepath(address, output_dir, ssh_client)
    LOG.info(f'starting iperf3 client process to > {address} , '
             f'output file is : {output_path}')
    # just in case there is some leftover file from previous run,
    # it needs to be removed, otherwise iperf will append new log
    # to the end of the existing file and this will make json output
    # file to be malformed
    files.remove_old_logfile(output_path, ssh_client=ssh_client)
    # If there is ssh client for the server where iperf3 server is going
    # to run, lets make sure it is started fresh as e.g. in case of
    # failure in the previous run, it may report that is still "busy" thus
    # iperf3 client will not start properly
    if iperf3_server_ssh_client:
        _stop_iperf3_server(
            port=port, protocol=protocol,
            ssh_client=iperf3_server_ssh_client)
        start_iperf3_server(
            port=port, protocol=protocol,
            ssh_client=iperf3_server_ssh_client)

        if not _iperf3_server_alive(
                port=port, protocol=protocol,
                ssh_client=iperf3_server_ssh_client):
            testcase = tobiko.get_test_case()
            testcase.fail('iperf3 server did not start properly '
                          f'on the server {iperf3_server_ssh_client}')

    # Now, finally iperf3 client should be ready to start
    execute_iperf3_client(
        address=address,
        bitrate=bitrate,
        download=download,
        port=port,
        protocol=protocol,
        ssh_client=ssh_client,
        logfile=output_path,
        run_in_background=True)


def _get_iperf3_pid(
        address: typing.Union[str, netaddr.IPAddress, None] = None,
        port: int = None,
        protocol: str = None,
        ssh_client: ssh.SSHClientType = None) -> typing.Union[int, None]:
    if address:
        iperf_commands = [f'iperf3 .*{address}']
    elif protocol and protocol.lower() == 'udp':
        iperf_commands = [f'iperf3 .*-s .*-u .*-p {port}',
                          f'iperf3 .*-s .*-p {port} .*-u']
    else:
        iperf_commands = [f'iperf3 .*-s .*-p {port}']

    for iperf_command in iperf_commands:
        iperf_processes = sh.list_processes(command_line=iperf_command,
                                            ssh_client=ssh_client)
        if iperf_processes:
            return iperf_processes.unique.pid
    LOG.debug('no iperf3 processes were found')
    return None


def _get_iperf3_log_raw(logfile: str,
                        ssh_client: ssh.SSHClientType = None):
    for attempt in tobiko.retry(timeout=60, interval=5):
        # download the iperf results file to local
        if ssh_client is not None:
            local_tmp_file = sh.local_shell_connection().make_temp_file()
            try:
                sh.get_file(logfile, local_tmp_file, ssh_client)
            except FileNotFoundError as err:
                message = f'Failed to download iperf log file. Error {err}'
                if attempt.is_last:
                    tobiko.fail(message)
                else:
                    LOG.debug(message)
                    LOG.debug('Retrying to download iperf log file...')
                    continue
            local_logfile = local_tmp_file
        else:
            local_logfile = logfile

        # this command has to be run locally
        iperf_log_raw = sh.execute(f"cat {local_logfile}").stdout

        iperf_log_raw = remove_log_lines_end_json_str(iperf_log_raw)
        LOG.debug(f'iperf log raw: {iperf_log_raw} ')

        # return if iperf_log_raw is not empty or if this is
        # called after the creation of the iperf background process
        if iperf_log_raw:
            return iperf_log_raw
        if not config.is_prevent_create():
            LOG.debug('iperf log file empty, which is normal after background '
                      'process creation')
            return  # return None

        if attempt.is_last:
            tobiko.fail('Failed empty iperf file.')


def parse_json_stream_output(iperf_log_raw):
    # Logs are printed by iperf3 client to the stdout or file in json
    # format, but the format is different than what is stored
    # without "--json-stream" option
    # So to be able to validate them in the same way, logs
    # need to be converted
    iperf3_results_data: dict = {
        "intervals": []
    }
    for log_line in iperf_log_raw.splitlines():
        log_line_json = json.loads(log_line)
        if log_line_json.get('event') != 'interval':
            continue
        iperf3_results_data["intervals"].append(
            log_line_json["data"])

    return iperf3_results_data


def check_iperf3_client_results(address: typing.Union[str, netaddr.IPAddress],
                                output_dir: str = 'tobiko_iperf_results',
                                ssh_client: ssh.SSHClientType = None,
                                **kwargs):  # noqa; pylint: disable=W0613
    logfile = get_iperf3_logs_filepath(address, output_dir, ssh_client)
    iperf_log_raw = _get_iperf3_log_raw(logfile, ssh_client)
    if not iperf_log_raw and not config.is_prevent_create():
        LOG.debug('empty iperf log file is ok when TOBIKO_PREVENT_CREATE is '
                  'disabled')
        return

    try:
        iperf_log = json.loads(iperf_log_raw)
    except json.JSONDecodeError:
        iperf_log = parse_json_stream_output(iperf_log_raw)
    longest_break = 0  # seconds
    breaks_total = 0  # seconds
    current_break = 0  # seconds
    intervals = iperf_log.get("intervals")
    if not intervals:
        tobiko.fail(f"No intervals data found in {logfile}")
    for interval in intervals:
        if interval["sum"]["bytes"] == 0:
            interval_duration = (
                interval["sum"]["end"] - interval["sum"]["start"])
            current_break += interval_duration
            if current_break > longest_break:
                longest_break = current_break
            breaks_total += interval_duration
        else:
            current_break = 0

    files.truncate_client_logfile(logfile, ssh_client)

    testcase = tobiko.get_test_case()
    testcase.assertLessEqual(longest_break,
                             CONF.tobiko.rhosp.max_traffic_break_allowed)
    testcase.assertLessEqual(breaks_total,
                             CONF.tobiko.rhosp.max_total_breaks_allowed)


def remove_log_lines_end_json_str(json_str: str) -> str:
    lines = json_str.splitlines()
    while lines:
        last_line = lines[-1].strip()
        if len(last_line) > 0 and last_line[-1] == "}":
            # Stop when we find }
            break
        # Remove last line, remove possible error logs
        lines.pop()
    return "\n".join(lines)


def iperf3_client_alive(address: typing.Union[str, netaddr.IPAddress],  # noqa; pylint: disable=W0613
                        ssh_client: ssh.SSHClientType = None,
                        **kwargs) -> bool:
    return bool(_get_iperf3_pid(address=address, ssh_client=ssh_client))


def stop_iperf3_client(address: typing.Union[str, netaddr.IPAddress],
                       ssh_client: ssh.SSHClientType = None,
                       **kwargs):  # noqa; pylint: disable=W0613
    pid = _get_iperf3_pid(address=address, ssh_client=ssh_client)
    if pid:
        LOG.info(f'iperf3 client process to > {address} already running '
                 f'with PID: {pid}')
        sh.execute(f'kill {pid}', ssh_client=ssh_client, sudo=True)
        # wait until iperf client process disappears
        sh.wait_for_processes(timeout=120,
                              sleep_interval=5,
                              ssh_client=ssh_client,
                              pid=pid)


def start_iperf3_server(
        port: typing.Union[int, None],
        protocol: typing.Union[str, None],
        ssh_client: ssh.SSHClientType):
    parameters = _parameters.iperf3_server_parameters(
        port=port, protocol=protocol)
    command = _interface.get_iperf3_server_command(parameters)
    process = sh.process(command, ssh_client=ssh_client)
    process.execute()


def _iperf3_server_alive(
        port: typing.Union[int, None],
        protocol: typing.Union[str, None],
        ssh_client: ssh.SSHClientType = None) -> bool:
    return bool(
        _get_iperf3_pid(port=port, protocol=protocol,
                        ssh_client=ssh_client))


def _stop_iperf3_server(
        port: typing.Union[int, None],
        protocol: typing.Union[str, None],
        ssh_client: ssh.SSHClientType = None):
    pid = _get_iperf3_pid(port=port, protocol=protocol, ssh_client=ssh_client)
    if pid:
        LOG.info(f'iperf3 server listening on the {protocol} port: {port} '
                 f'is already running with PID: {pid}')
        sh.execute(f'sudo kill {pid}', ssh_client=ssh_client)
        # wait until iperf server process disappears
        sh.wait_for_processes(timeout=120,
                              sleep_interval=5,
                              ssh_client=ssh_client,
                              pid=pid)



# Copyright (c) 2025 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

import netaddr

from tobiko.tripleo import overcloud
from tobiko.shell import iperf3
from tobiko.shell import ping
from tobiko.shell import sh
from tobiko.shell import ssh


# Test is inteded for D/S env
@overcloud.skip_if_missing_overcloud
def check_or_start_background_vm_ping(
        server_ip: typing.Union[str, netaddr.IPAddress],
        ssh_client: ssh.SSHClientType = None):
    """Check if process exists, if so stop and check ping health
    if not : start a new separate ping process.
    Executes a Background ping to a vm floating_ip,
    this test is intended to be run and picked up again
    by the next tobiko run. Ping results are parsed
    and a failure is raised if ping failure is above a certain amount"""
    if ssh_client is None:
        sh.check_or_start_background_process(
            bg_function=ping.write_ping_to_file,
            bg_process_name='tobiko_background_ping',
            check_function=ping.check_ping_statistics,
            ping_ip=server_ip)
    else:
        sh.check_or_start_external_process(
            start_function=ping.execute_ping_in_background,
            check_function=ping.check_ping_results,
            liveness_function=ping.ping_alive,
            stop_function=ping.stop_ping,
            address=server_ip,
            ssh_client=ssh_client)


# Test is inteded for D/S env
@overcloud.skip_if_missing_overcloud
def check_or_start_background_iperf_connection(
        server_ip: typing.Union[str, netaddr.IPAddress],
        port: int,
        protocol: str,
        ssh_client: ssh.SSHClientType = None,
        iperf3_server_ssh_client: ssh.SSHClientType = None):
    """Check if process exists, if so stop and check ping health
    if not : start a new separate iperf client process.
    Iperf server runs on the vm which is behind IP address given
    as the 'server_ip'.
    If `iperf3_server_ssh_client` is given, Tobiko will make sure
    that iperf3 server is running on the server behind this ssh_client
    """
    sh.check_or_start_external_process(
        start_function=iperf3.execute_iperf3_client_in_background,
        check_function=iperf3.check_iperf3_client_results,
        liveness_function=iperf3.iperf3_client_alive,
        stop_function=iperf3.stop_iperf3_client,
        address=server_ip,
        port=port,
        protocol=protocol,
        ssh_client=ssh_client,
        iperf3_server_ssh_client=iperf3_server_ssh_client)



# Copyright (c) 2025 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

from oslo_log import log

from tobiko.shell.sh import _execute
from tobiko.shell import ssh


LOG = log.getLogger(__name__)


def get_nm_connection_ids(ssh_client: ssh.SSHClientType = None) -> list:
    result = _execute.execute('nmcli -g UUID con',
                              ssh_client=ssh_client)
    return result.stdout.splitlines()


def get_nm_connection_values(connection: str,
                             values: str,
                             ssh_client: ssh.SSHClientType = None) -> list:
    result = _execute.execute(f'nmcli -g {values} con show "{connection}"',
                              ssh_client=ssh_client)
    return_values = []
    for line in result.stdout.splitlines():
        if line:
            for value in line.split('|'):
                # nmcli adds escape char before ":" and we need to remove it
                return_values.append(value.strip().replace('\\', ''))

    return return_values



# Copyright (c) 2025 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from oslo_log import log

import tobiko
from tobiko.openstack.nova import _client
from tobiko.openstack.nova import _server
from tobiko.openstack.nova import _service
from tobiko.openstack import topology
from tobiko.shell import ping
from tobiko.shell import sh


LOG = log.getLogger(__name__)


def check_nova_services_health(timeout=600., interval=2.):
    retry = tobiko.retry(timeout=timeout, interval=interval)
    _service.wait_for_services_up(retry=retry)


def check_virsh_domains_running():
    """check all vms are running via virsh list command"""
    for compute in topology.list_openstack_nodes(group='compute'):
        hostname = sh.get_hostname(ssh_client=compute.ssh_client,
                                   fqdn=True)
        param = {'OS-EXT-SRV-ATTR:hypervisor_hostname': hostname}
        vm_list_per_compute = _client.list_servers(**param)
        for vm in vm_list_per_compute:
            for attempt in tobiko.retry(timeout=120, interval=5):
                if check_vm_running_via_virsh(compute, vm.id):
                    LOG.info(f"{vm.id} is running ok on {hostname}")
                    break
                else:
                    msg = f"{vm.id} is not in running state on {hostname}"
                    if attempt.is_last:
                        tobiko.fail("timeout!! " + msg)
                    LOG.error(f"{vm.id} is not in running state on "
                              f"{hostname} ... Retrying")


def check_vms_ping(vm_list):
    for vm in vm_list:
        fip = _server.list_server_ip_addresses(vm,
                                               address_type='floating').first
        ping.ping_until_received(fip).assert_replied()


def check_vm_evacuations(vms_old=None, compute_host=None, timeout=600,
                         interval=2, check_no_evacuation=False):
    """check evacuation of vms
        'vms_old' - all the vms that originally (before evacuation) were
        running on the 'compute_host' machine"""
    for attempt in tobiko.retry(timeout=timeout, interval=interval):
        failures = []
        vms_new = _client.list_servers()
        for vm_old in vms_old or []:
            vm_evacuated = vms_new.with_attributes(  # pylint: disable=W0212
                id=vm_old.id).unique
            new_vm_host = vm_evacuated._info[  # pylint: disable=W0212
                'OS-EXT-SRV-ATTR:hypervisor_hostname']
            LOG.info(f"server {vm_old} evacuated to {new_vm_host}")

            if check_no_evacuation:
                cond = bool(compute_host != new_vm_host)
            else:
                cond = bool(compute_host == new_vm_host)
            if cond:
                failures.append(
                    'Failed vm evacuations: {}\n\n'.format(vm_old))

        if not failures:
            LOG.debug(vms_old)
            LOG.debug('All vms were evacuated!')
            return

        if attempt.is_last:
            tobiko.fail(
                'Timeout checking VM evacuations:\n{!s}', '\n'.join(failures))
        else:
            LOG.error('Failed nova evacuation:\n {}'.format(failures))
            LOG.error('Retrying...')


def check_vm_running_via_virsh(topology_compute, vm_id):
    """check that a vm is in running state via virsh command,
    return false if not"""
    if vm_id in get_vm_uuid_list_running_via_virsh(topology_compute):
        return True
    else:
        return False


def get_vm_uuid_list_running_via_virsh(topology_compute):
    from tobiko import podified
    from tobiko.tripleo import containers
    from tobiko.tripleo import overcloud

    get_uuid_loop = ("for i in `virsh list --name --state-running`; do "
                     "virsh domuuid $i; done")
    containerized_libvirt_cmd = \
        "{container_runtime} exec -u root {nova_libvirt} sh -c '{get_uuids}'"

    if podified.has_podified_cp():
        command = containerized_libvirt_cmd.format(
            container_runtime=podified.CONTAINER_RUNTIME,
            nova_libvirt=podified.NOVA_LIBVIRT_CONTAINER,
            get_uuids=get_uuid_loop)
    elif overcloud.has_overcloud():
        command = containerized_libvirt_cmd.format(
            container_runtime=containers.get_container_runtime_name(),
            nova_libvirt=containers.get_libvirt_container_name(),
            get_uuids=get_uuid_loop)
    else:
        command = get_uuid_loop

    return sh.execute(command,
                      ssh_client=topology_compute.ssh_client,
                      sudo=True).stdout.split()


def wait_for_all_instances_status(status, timeout=None):
    """wait for all instances for a certain status or raise an exception"""
    for instance in _client.list_servers():
        _client.wait_for_server_status(server=instance.id, status=status,
                                       timeout=timeout)
        instance_info = 'instance {nova_instance} is {state} on {host}'.format(
            nova_instance=instance.name,
            state=status,
            host=instance._info[  # pylint: disable=W0212
                'OS-EXT-SRV-ATTR:hypervisor_hostname'])
        LOG.info(instance_info)



# Copyright (c) 2025 Red Hat, Inc.
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from __future__ import absolute_import

import typing

from manilaclient.v2.shares import Share
from oslo_log import log

import tobiko
from tobiko.openstack import manila
from tobiko.openstack.base import _fixture as base_fixture

LOG = log.getLogger(__name__)


class ManilaShareFixture(base_fixture.ResourceFixture):

    _resource: typing.Optional[Share] = None
    share_protocol: typing.Optional[str] = None
    size: typing.Optional[int] = None

    def __init__(self, share_protocol=None, size=None):
        super().__init__()
        self.share_protocol = share_protocol or self.share_protocol
        self.size = size or self.size

    @property
    def share_id(self):
        return self.resource_id

    @property
    def share(self):
        return self.resource

    @tobiko.interworker_synched('manila_setup_fixture')
    def try_create_resource(self):
        super().try_create_resource()

    def resource_create(self):
        share = manila.create_share(share_protocol=self.share_protocol,
                                    size=self.size,
                                    name=self.name)
        manila.wait_for_share_status(share['id'])
        LOG.debug(f'Share {share["name"]} was deployed successfully '
                  f'with id {share["id"]}')
        return share

    def resource_delete(self):
        LOG.debug('Deleting Share %r ...', self.name)
        manila.delete_share(self.share_id)
        manila.wait_for_resource_deletion(self.share_id)
        LOG.debug('Share %r deleted.', self.name)

    def resource_find(self):
        found_shares = manila.get_shares_by_name(self.name)
        if len(found_shares) > 1:
            tobiko.fail(f'Unexpected number of shares found: {found_shares}')

        if found_shares:
            LOG.debug("Share %r found.", self.name)
            return found_shares[0]

        # no shares found
        LOG.debug("Share %r not found.", self.name)
